import pytest

from spacy.tokens import Doc


# fmt: off
@pytest.mark.parametrize(
    "words,heads,deps,pos,chunk_offsets",
    [
        # un gato -> "un gato"
        (
            ["un", "gato"],
            [1, 1],
            ["det", "ROOT"],
            ["DET", "NOUN"],
            [(0, 2)],
        ),
        # la camisa negra -> "la camisa negra"
        (
            ["la", "camisa", "negra"],
            [1, 1, 1],
            ["det", "ROOT", "amod"],
            ["DET", "NOUN", "ADJ"],
            [(0, 3)],
        ),
        # un lindo gatito -> "un lindo gatito"
        (
            ["Un", "lindo", "gatito"],
            [2, 2, 2],
            ["det", "amod", "ROOT"],
            ["DET", "ADJ", "NOUN"],
            [(0,3)]
        ),
        # una chica hermosa e inteligente -> una chica hermosa e inteligente
        (
            ["Una", "chica", "hermosa", "e", "inteligente"],
            [1, 1, 1, 4, 2],
            ["det", "ROOT", "amod", "cc", "conj"],
            ["DET", "NOUN", "ADJ", "CCONJ", "ADJ"],
            [(0,5)]
        ),
        # el fabuloso gato pardo -> "el fabuloso gato pardo"
        (
            ["el", "fabuloso", "gato", "pardo"],
            [2, 2, 2, 2],
            ["det", "amod", "ROOT", "amod"],
            ["DET", "ADJ", "NOUN", "ADJ"],
            [(0,4)]
        ),
        # Tengo un gato y un perro -> un gato, un perro
        ( 
            ["Tengo", "un", "gato", "y", "un", "perro"],
            [0, 2, 0, 5, 5, 0],
            ["ROOT", "det", "obj", "cc", "det", "conj"],
            ["VERB", "DET", "NOUN", "CCONJ", "DET", "NOUN"],
            [(1,3), (4,6)]
         
        ),
        # Dom Pedro II -> Dom Pedro II
        (
            ["Dom", "Pedro", "II"],
            [0, 0, 0],
            ["ROOT", "flat", "flat"],
            ["PROPN", "PROPN", "PROPN"],
            [(0,3)]
        ),
        # los Estados Unidos -> los Estados Unidos
        (
            ["los", "Estados", "Unidos"],
            [1, 1, 1],
            ["det", "ROOT", "flat"],
            ["DET", "PROPN", "PROPN"],
            [(0,3)]
        ),
        # Miguel de Cervantes -> Miguel de Cervantes
        (
            ["Miguel", "de", "Cervantes"],
            [0, 2, 0],
            ["ROOT", "case", "flat"],
            ["PROPN", "ADP", "PROPN"],
            [(0,3)]
        ),
        (
            ["Rio", "de", "Janeiro"],
            [0, 2, 0],
            ["ROOT", "case", "flat"],
            ["PROPN", "ADP", "PROPN"],
            [(0,3)]
        ),
        # la destrucción de la ciudad -> la destrucción, la ciudad
        (
            ["la", "destrucción", "de", "la", "ciudad"],
            [1, 1, 4, 4, 1],
            ['det', 'ROOT', 'case', 'det', 'nmod'],
            ['DET', 'NOUN', 'ADP', 'DET', 'NOUN'],
            [(0,2), (3,5)]
        ),
        # la traducción de Susana del informe -> la traducción, Susana, informe
        (
            ['la', 'traducción', 'de', 'Susana', 'del', 'informe'],
            [1, 1, 3, 1, 5, 1],
            ['det', 'ROOT', 'case', 'nmod', 'case', 'nmod'],
            ['DET', 'NOUN', 'ADP', 'PROPN', 'ADP', 'NOUN'],
            [(0,2), (3,4), (5,6)]  
       
        ),
        # El gato regordete de Susana y su amigo -> el gato regordete, Susana, su amigo
        (  
            ['El', 'gato', 'regordete', 'de', 'Susana', 'y', 'su', 'amigo'],
            [1, 1, 1, 4, 1, 7, 7, 1],
            ['det', 'ROOT', 'amod', 'case', 'nmod', 'cc', 'det', 'conj'],
            ['DET', 'NOUN', 'ADJ', 'ADP', 'PROPN', 'CCONJ', 'DET', 'NOUN'],
            [(0,3), (4,5), (6,8)]
        ),
        # Afirmó que sigue el criterio europeo y que trata de incentivar el mercado donde no lo hay -> el criterio europeo, el mercado, donde, lo
        (
            ['Afirmó', 'que', 'sigue', 'el', 'criterio', 'europeo', 'y', 'que', 'trata', 'de', 'incentivar', 'el', 'mercado', 'donde', 'no', 'lo', 'hay'],
            [0, 2, 0, 4, 2, 4, 8, 8, 2, 10, 8, 12, 10, 16, 16, 16, 0],
            ['ROOT', 'mark', 'ccomp', 'det', 'obj', 'amod', 'cc', 'mark', 'conj', 'mark', 'xcomp', 'det', 'obj', 'obl', 'advmod', 'obj', 'advcl'],
            ['VERB', 'SCONJ', 'VERB', 'DET', 'NOUN', 'ADJ', 'CCONJ', 'SCONJ', 'VERB', 'ADP', 'VERB', 'DET', 'NOUN', 'PRON', 'ADV', 'PRON', 'AUX'],
            [(3,6), (11,13), (13,14), (15,16)]
        ),
        # En este sentido se refirió a la reciente creación del Ministerio de Ciencia y Tecnología y a las primeras declaraciones de su titular, Anna Birulés, sobre el impulso de la investigación, desarrollo e innovación -> este sentido, se, la reciente creación, Ministerio de Ciencia y Tecnología, a las primeras declaraciones, su titular, , Anna Birulés,, el impulso, la investigación, , desarrollo, innovación
        (
            ['En', 'este', 'sentido', 'se', 'refirió', 'a', 'la', 'reciente', 'creación', 'del', 'Ministerio', 'de', 'Ciencia', 'y', 'Tecnología', 'y', 'a', 'las', 'primeras', 'declaraciones', 'de', 'su', 'titular', ',', 'Anna', 'Birulés', ',', 'sobre', 'el', 'impulso', 'de', 'la', 'investigación', ',', 'desarrollo', 'e', 'innovación'],
            [2, 2, 4, 4, 4, 8, 8, 8, 4, 10, 8, 12, 10, 14, 12, 19, 19, 19, 19, 8, 22, 22, 19, 24, 22, 24, 24, 29, 29, 19, 32, 32, 29, 34, 32, 36, 32],
            ['case', 'det', 'obl', 'obj', 'ROOT', 'case', 'det', 'amod', 'obj', 'case', 'nmod', 'case', 'flat', 'cc', 'conj', 'cc', 'case', 'det', 'amod', 'conj', 'case', 'det', 'nmod', 'punct', 'appos', 'flat', 'punct', 'case', 'det', 'nmod', 'case', 'det', 'nmod', 'punct', 'conj', 'cc', 'conj'],
            ['ADP', 'DET', 'NOUN', 'PRON', 'VERB', 'ADP', 'DET', 'ADJ', 'NOUN', 'ADP', 'PROPN', 'ADP', 'PROPN', 'CCONJ', 'PROPN', 'CCONJ', 'ADP', 'DET', 'ADJ', 'NOUN', 'ADP', 'DET', 'NOUN', 'PUNCT', 'PROPN', 'PROPN', 'PUNCT', 'ADP', 'DET', 'NOUN', 'ADP', 'DET', 'NOUN', 'PUNCT', 'NOUN', 'CCONJ', 'NOUN'],
            [(1, 3), (3, 4), (6, 9), (10, 15), (16, 20), (21, 23), (23, 27), (28, 30), (31, 33), (33, 35), (36, 37)]
        ),
        # Asimismo defiende la financiación pública de la investigación básica y pone de manifiesto que las empresas se centran más en la investigación y desarrollo con objetivos de mercado. -> la financiación pública, la investigación básica, manifiesto, las empresas, se, la investigación, desarrollo, objetivos, mercado
        (
            ['Asimismo', 'defiende', 'la', 'financiación', 'pública', 'de', 'la', 'investigación', 'básica', 'y', 'pone', 'de', 'manifiesto', 'que', 'las', 'empresas', 'se', 'centran', 'más', 'en', 'la', 'investigación', 'y', 'desarrollo', 'con', 'objetivos', 'de', 'mercado'],
            [1, 1, 3, 1, 3, 7, 7, 3, 7, 10, 1, 12, 10, 17, 15, 17, 17, 10, 17, 21, 21, 17, 23, 21, 25, 17, 27, 25],
            ['advmod', 'ROOT', 'det', 'obj', 'amod', 'case', 'det', 'nmod', 'amod', 'cc', 'conj', 'case', 'obl', 'mark', 'det', 'nsubj', 'obj', 'ccomp', 'obj', 'case', 'det', 'obl', 'cc', 'conj', 'case', 'obl', 'case', 'nmod'],
            ['ADV', 'VERB', 'DET', 'NOUN', 'ADJ', 'ADP', 'DET', 'NOUN', 'ADJ', 'CCONJ', 'VERB', 'ADP', 'NOUN', 'SCONJ', 'DET', 'NOUN', 'PRON', 'VERB', 'ADV', 'ADP', 'DET', 'NOUN', 'CCONJ', 'NOUN', 'ADP', 'NOUN', 'ADP', 'NOUN'],
            [(2, 5), (6, 9), (12, 13), (14, 16), (16, 17), (20, 22), (23, 24), (25, 26), (27, 28)]
        ),
        # Tras indicar que la inversión media en investigación en la Unión Europea se sitúa en el 1,8 por ciento del PIB, frente al 2,8 por ciento en Japón y EEUU, Couceiro dijo que España está en "el buen camino" y se está creando un entorno propicio para la innovación empresarial' -> la inversión media, investigación, la Unión Europea, se, PIB, Japón, EEUU, Couceiro, España, se, un entorno propicio para la innovación empresaria
        (
            ['Tras', 'indicar', 'que', 'la', 'inversión', 'media', 'en', 'investigación', 'en', 'la', 'Unión', 'Europea', 'se', 'sitúa', 'en', 'el', '1,8', 'por', 'ciento', 'del', 'PIB', ',', 'frente', 'al', '2,8', 'por', 'ciento', 'en', 'Japón', 'y', 'EEUU', ',', 'Couceiro', 'dijo', 'que', 'España', 'está', 'en', '"', 'el', 'buen', 'camino', '"', 'y', 'se', 'está', 'creando', 'un', 'entorno', 'propicio', 'para', 'la', 'innovación', 'empresarial'],
            [1, 33, 13, 4, 13, 4, 7, 4, 10, 10, 4, 10, 13, 1, 16, 16, 13, 18, 16, 20, 16, 24, 24, 22, 13, 26, 24, 28, 24, 30, 28, 1, 33, 33, 41, 41, 41, 41, 41, 41, 41, 33, 41, 46, 46, 46, 33, 48, 46, 48, 52, 52, 49, 52],
            ['mark', 'advcl', 'mark', 'det', 'nsubj', 'amod', 'case', 'nmod', 'case', 'det', 'nmod', 'flat', 'obj', 'ccomp', 'case', 'det', 'obj', 'case', 'compound', 'case', 'nmod', 'punct', 'case', 'fixed', 'obl', 'case', 'compound', 'case', 'nmod', 'cc', 'conj', 'punct', 'nsubj', 'ROOT', 'mark', 'nsubj', 'cop', 'case', 'punct', 'det', 'amod', 'ccomp', 'punct', 'cc', 'obj', 'aux', 'conj', 'det', 'nsubj', 'amod', 'case', 'det', 'nmod', 'amod'],
            ['ADP', 'VERB', 'SCONJ', 'DET', 'NOUN', 'ADJ', 'ADP', 'NOUN', 'ADP', 'DET', 'PROPN', 'PROPN', 'PRON', 'VERB', 'ADP', 'DET', 'NUM', 'ADP', 'NUM', 'ADP', 'PROPN', 'PUNCT', 'NOUN', 'ADP', 'NUM', 'ADP', 'NUM', 'ADP', 'PROPN', 'CCONJ', 'PROPN', 'PUNCT', 'PROPN', 'VERB', 'SCONJ', 'PROPN', 'AUX', 'ADP', 'PUNCT', 'DET', 'ADJ', 'NOUN', 'PUNCT', 'CCONJ', 'PRON', 'AUX', 'VERB', 'DET', 'NOUN', 'ADJ', 'ADP', 'DET', 'NOUN', 'ADJ'],
            [(3, 6), (7, 8), (9, 12), (12, 13), (20, 21), (28, 29), (30, 31), (32, 33), (35, 36), (44, 45), (47, 54)]
        ),
    ],
)
# fmt: on
def test_es_noun_chunks(es_vocab, words, heads, deps, pos, chunk_offsets):
    doc = Doc(es_vocab, words=words, heads=heads, deps=deps, pos=pos)
    assert [(c.start, c.end) for c in doc.noun_chunks] == chunk_offsets


def test_noun_chunks_is_parsed_es(es_tokenizer):
    """Test that noun_chunks raises Value Error for 'es' language if Doc is not parsed."""
    doc = es_tokenizer("en Oxford este verano")
    with pytest.raises(ValueError):
        list(doc.noun_chunks)



from cymem.cymem cimport Pool

from ...strings cimport StringStore
from ...structs cimport TokenC
from ...training.example cimport Example
from ...typedefs cimport attr_t, weight_t
from ._state cimport StateC
from .stateclass cimport StateClass


cdef struct Transition:
    int clas
    int move
    attr_t label

    weight_t score

    bint (*is_valid)(const StateC* state, attr_t label) nogil
    weight_t (*get_cost)(const StateC* state, const void* gold, attr_t label) nogil
    int (*do)(StateC* state, attr_t label) nogil


ctypedef weight_t (*get_cost_func_t)(
    const StateC* state, const void* gold, attr_tlabel
) nogil
ctypedef weight_t (*move_cost_func_t)(
        const StateC* state, const void* gold
) nogil
ctypedef weight_t (*label_cost_func_t)(
    const StateC* state, const void* gold, attr_t label
) nogil

ctypedef int (*do_func_t)(StateC* state, attr_t label) nogil

ctypedef void* (*init_state_t)(Pool mem, int length, void* tokens) except NULL

ctypedef int (*del_state_t)(Pool mem, void* state, void* extra_args) except -1

cdef class TransitionSystem:
    cdef Pool mem
    cdef StringStore strings
    cdef Transition* c
    cdef readonly int n_moves
    cdef int _size
    cdef public attr_t root_label
    cdef public freqs
    cdef public object labels
    cdef public object cfg
    cdef init_state_t init_beam_state
    cdef del_state_t del_beam_state

    cdef Transition lookup_transition(self, object name) except *

    cdef Transition init_transition(self, int clas, int move, attr_t label) except *

    cdef int set_valid(self, int* output, const StateC* st) nogil

    cdef int set_costs(self, int* is_valid, weight_t* costs,
                       const StateC* state, gold) except -1



"""
Example sentences to test spaCy and its language models.
>>> from spacy.lang.tt.examples import sentences
>>> docs = nlp.pipe(sentences)
"""

sentences = [
    "Apple Бөекбритания стартабын $1 миллиард өчен сатып алыун исәпли.",
    "Автоном автомобильләр иминият җаваплылыкны җитештерүчеләргә күчерә.",
    "Сан-Франциско тротуар буенча йөри торган робот-курьерларны тыю мөмкинлеген карый.",
    "Лондон - Бөекбританиядә урнашкан зур шәһәр.",
    "Син кайда?",
    "Францияда кем президент?",
    "Америка Кушма Штатларының башкаласы нинди шәһәр?",
    "Барак Обама кайчан туган?",
]



import hashlib
import os
import shutil
import sys
from configparser import InterpolationError
from contextlib import contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    overload,
)

import srsly
import typer
from click import NoSuchOption
from click.parser import split_arg_string
from thinc.api import Config, ConfigValidationError, require_gpu
from thinc.util import gpu_is_available
from typer.main import get_command
from wasabi import Printer, msg
from weasel import app as project_cli

from .. import about
from ..compat import Literal
from ..schemas import validate
from ..util import (
    ENV_VARS,
    SimpleFrozenDict,
    import_file,
    is_compatible_version,
    logger,
    make_tempdir,
    registry,
    run_command,
)

SDIST_SUFFIX = ".tar.gz"
WHEEL_SUFFIX = "-py3-none-any.whl"

PROJECT_FILE = "project.yml"
COMMAND = "python -m spacy"
NAME = "spacy"
HELP = """spaCy Command-line Interface

DOCS: https://spacy.io/api/cli
"""
PROJECT_HELP = f"""Command-line interface for spaCy projects and templates.
You'd typically start by cloning a project template to a local directory and
fetching its assets like datasets etc. See the project's {PROJECT_FILE} for the
available commands.
"""
DEBUG_HELP = """Suite of helpful commands for debugging and profiling. Includes
commands to check and validate your config files, training and evaluation data,
and custom model implementations.
"""
BENCHMARK_HELP = """Commands for benchmarking pipelines."""
INIT_HELP = """Commands for initializing configs and pipeline packages."""

# Wrappers for Typer's annotations. Initially created to set defaults and to
# keep the names short, but not needed at the moment.
Arg = typer.Argument
Opt = typer.Option

app = typer.Typer(name=NAME, help=HELP)
benchmark_cli = typer.Typer(name="benchmark", help=BENCHMARK_HELP, no_args_is_help=True)
debug_cli = typer.Typer(name="debug", help=DEBUG_HELP, no_args_is_help=True)
init_cli = typer.Typer(name="init", help=INIT_HELP, no_args_is_help=True)

app.add_typer(project_cli, name="project", help=PROJECT_HELP, no_args_is_help=True)
app.add_typer(debug_cli)
app.add_typer(benchmark_cli)
app.add_typer(init_cli)


def setup_cli() -> None:
    # Make sure the entry-point for CLI runs, so that they get imported.
    registry.cli.get_all()
    # Ensure that the help messages always display the correct prompt
    command = get_command(app)
    command(prog_name=COMMAND)


def parse_config_overrides(
    args: List[str], env_var: Optional[str] = ENV_VARS.CONFIG_OVERRIDES
) -> Dict[str, Any]:
    """Generate a dictionary of config overrides based on the extra arguments
    provided on the CLI, e.g. --training.batch_size to override
    "training.batch_size". Arguments without a "." are considered invalid,
    since the config only allows top-level sections to exist.

    env_vars (Optional[str]): Optional environment variable to read from.
    RETURNS (Dict[str, Any]): The parsed dict, keyed by nested config setting.
    """
    env_string = os.environ.get(env_var, "") if env_var else ""
    env_overrides = _parse_overrides(split_arg_string(env_string))
    cli_overrides = _parse_overrides(args, is_cli=True)
    if cli_overrides:
        keys = [k for k in cli_overrides if k not in env_overrides]
        logger.debug("Config overrides from CLI: %s", keys)
    if env_overrides:
        logger.debug("Config overrides from env variables: %s", list(env_overrides))
    return {**cli_overrides, **env_overrides}


def _parse_overrides(args: List[str], is_cli: bool = False) -> Dict[str, Any]:
    result = {}
    while args:
        opt = args.pop(0)
        err = f"Invalid config override '{opt}'"
        if opt.startswith("--"):  # new argument
            orig_opt = opt
            opt = opt.replace("--", "")
            if "." not in opt:
                if is_cli:
                    raise NoSuchOption(orig_opt)
                else:
                    msg.fail(f"{err}: can't override top-level sections", exits=1)
            if "=" in opt:  # we have --opt=value
                opt, value = opt.split("=", 1)
                opt = opt.replace("-", "_")
            else:
                if not args or args[0].startswith("--"):  # flag with no value
                    value = "true"
                else:
                    value = args.pop(0)
            result[opt] = _parse_override(value)
        else:
            msg.fail(f"{err}: name should start with --", exits=1)
    return result


def _parse_override(value: Any) -> Any:
    # Just like we do in the config, we're calling json.loads on the
    # values. But since they come from the CLI, it'd be unintuitive to
    # explicitly mark strings with escaped quotes. So we're working
    # around that here by falling back to a string if parsing fails.
    # TODO: improve logic to handle simple types like list of strings?
    try:
        return srsly.json_loads(value)
    except ValueError:
        return str(value)


@contextmanager
def show_validation_error(
    file_path: Optional[Union[str, Path]] = None,
    *,
    title: Optional[str] = None,
    desc: str = "",
    show_config: Optional[bool] = None,
    hint_fill: bool = True,
):
    """Helper to show custom config validation errors on the CLI.

    file_path (str / Path): Optional file path of config file, used in hints.
    title (str): Override title of custom formatted error.
    desc (str): Override description of custom formatted error.
    show_config (bool): Whether to output the config the error refers to.
    hint_fill (bool): Show hint about filling config.
    """
    try:
        yield
    except ConfigValidationError as e:
        title = title if title is not None else e.title
        if e.desc:
            desc = f"{e.desc}" if not desc else f"{e.desc}\n\n{desc}"
        # Re-generate a new error object with overrides
        err = e.from_error(e, title="", desc=desc, show_config=show_config)
        msg.fail(title)
        print(err.text.strip())
        if hint_fill and "value_error.missing" in err.error_types:
            config_path = (
                file_path
                if file_path is not None and str(file_path) != "-"
                else "config.cfg"
            )
            msg.text(
                "If your config contains missing values, you can run the 'init "
                "fill-config' command to fill in all the defaults, if possible:",
                spaced=True,
            )
            print(f"{COMMAND} init fill-config {config_path} {config_path} \n")
        sys.exit(1)
    except InterpolationError as e:
        msg.fail("Config validation error", e, exits=1)


def import_code(code_path: Optional[Union[Path, str]]) -> None:
    """Helper to import Python file provided in training commands / commands
    using the config. This makes custom registered functions available.
    """
    if code_path is not None:
        if not Path(code_path).exists():
            msg.fail("Path to Python code not found", code_path, exits=1)
        try:
            import_file("python_code", code_path)
        except Exception as e:
            msg.fail(f"Couldn't load Python code: {code_path}", e, exits=1)


def get_git_version(
    error: str = "Could not run 'git'. Make sure it's installed and the executable is available.",
) -> Tuple[int, int]:
    """Get the version of git and raise an error if calling 'git --version' fails.
    error (str): The error message to show.
    RETURNS (Tuple[int, int]): The version as a (major, minor) tuple. Returns
        (0, 0) if the version couldn't be determined.
    """
    try:
        ret = run_command("git --version", capture=True)
    except:
        raise RuntimeError(error)
    stdout = ret.stdout.strip()
    if not stdout or not stdout.startswith("git version"):
        return 0, 0
    version = stdout[11:].strip().split(".")
    return int(version[0]), int(version[1])


@overload
def string_to_list(value: str, intify: Literal[False] = ...) -> List[str]:
    ...


@overload
def string_to_list(value: str, intify: Literal[True]) -> List[int]:
    ...


def string_to_list(value: str, intify: bool = False) -> Union[List[str], List[int]]:
    """Parse a comma-separated string to a list and account for various
    formatting options. Mostly used to handle CLI arguments that take a list of
    comma-separated values.

    value (str): The value to parse.
    intify (bool): Whether to convert values to ints.
    RETURNS (Union[List[str], List[int]]): A list of strings or ints.
    """
    if not value:
        return []  # type: ignore[return-value]
    if value.startswith("[") and value.endswith("]"):
        value = value[1:-1]
    result = []
    for p in value.split(","):
        p = p.strip()
        if p.startswith("'") and p.endswith("'"):
            p = p[1:-1]
        if p.startswith('"') and p.endswith('"'):
            p = p[1:-1]
        p = p.strip()
        if intify:
            p = int(p)  # type: ignore[assignment]
        result.append(p)
    return result


def setup_gpu(use_gpu: int, silent=None) -> None:
    """Configure the GPU and log info."""
    if silent is None:
        local_msg = Printer()
    else:
        local_msg = Printer(no_print=silent, pretty=not silent)
    if use_gpu >= 0:
        local_msg.info(f"Using GPU: {use_gpu}")
        require_gpu(use_gpu)
    else:
        local_msg.info("Using CPU")
        if gpu_is_available():
            local_msg.info("To switch to GPU 0, use the option: --gpu-id 0")


def walk_directory(path: Path, suffix: Optional[str] = None) -> List[Path]:
    """Given a directory and a suffix, recursively find all files matching the suffix.
    Directories or files with names beginning with a . are ignored, but hidden flags on
    filesystems are not checked.
    When provided with a suffix `None`, there is no suffix-based filtering."""
    if not path.is_dir():
        return [path]
    paths = [path]
    locs = []
    seen = set()
    for path in paths:
        if str(path) in seen:
            continue
        seen.add(str(path))
        if path.parts[-1].startswith("."):
            continue
        elif path.is_dir():
            paths.extend(path.iterdir())
        elif suffix is not None and not path.parts[-1].endswith(suffix):
            continue
        else:
            locs.append(path)
    # It's good to sort these, in case the ordering messes up cache.
    locs.sort()
    return locs


def _format_number(number: Union[int, float], ndigits: int = 2) -> str:
    """Formats a number (float or int) rounding to `ndigits`, without truncating trailing 0s,
    as happens with `round(number, ndigits)`"""
    if isinstance(number, float):
        return f"{number:.{ndigits}f}"
    else:
        return str(number)



"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.hsb.examples import sentences
>>> docs = nlp.pipe(sentences)
"""


sentences = [
    "To běšo wjelgin raźone a jo se wót luźi derje pśiwzeło. Tak som dožywiła wjelgin",
    "Jogo pśewóźowarce stej groniłej, až how w serbskich stronach njama Santa Claus nic pytaś.",
    "A ten sobuźěłaśeŕ Statneje biblioteki w Barlinju jo pśimjeł drogotne knigły bźez rukajcowu z nagima rukoma!",
    "Take wobchadanje z našym kulturnym derbstwom zewšym njejźo.",
    "Wopśimjeśe drugich pśinoskow jo było na wusokem niwowje, ako pśecej.",
]



from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Optional,
    Tuple,
    Union,
)

from ..language import Language
from ..tokens.doc import Doc
from ..training import Example

class Pipe:
    def __call__(self, doc: Doc) -> Doc: ...
    def pipe(
        self, stream: Iterable[Doc], *, batch_size: int = ...
    ) -> Iterator[Doc]: ...
    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Language = ...,
    ) -> None: ...
    def score(
        self, examples: Iterable[Example], **kwargs: Any
    ) -> Dict[str, Union[float, Dict[str, float]]]: ...
    @property
    def is_trainable(self) -> bool: ...
    @property
    def labels(self) -> Tuple[str, ...]: ...
    @property
    def hide_labels(self) -> bool: ...
    @property
    def label_data(self) -> Any: ...
    def _require_labels(self) -> None: ...
    def set_error_handler(
        self, error_handler: Callable[[str, "Pipe", List[Doc], Exception], NoReturn]
    ) -> None: ...
    def get_error_handler(
        self,
    ) -> Callable[[str, "Pipe", List[Doc], Exception], NoReturn]: ...

def deserialize_config(path: Path) -> Any: ...



import pickle
import re

import pytest

from spacy.lang.en import English
from spacy.lang.it import Italian
from spacy.language import Language
from spacy.tokenizer import Tokenizer
from spacy.training import Example
from spacy.util import load_config_from_str

from ..util import make_tempdir


@pytest.fixture
def meta_data():
    return {
        "name": "name-in-fixture",
        "version": "version-in-fixture",
        "description": "description-in-fixture",
        "author": "author-in-fixture",
        "email": "email-in-fixture",
        "url": "url-in-fixture",
        "license": "license-in-fixture",
        "vectors": {"width": 0, "vectors": 0, "keys": 0, "name": None},
    }


@pytest.mark.issue(2482)
def test_issue2482():
    """Test we can serialize and deserialize a blank NER or parser model."""
    nlp = Italian()
    nlp.add_pipe("ner")
    b = nlp.to_bytes()
    Italian().from_bytes(b)


CONFIG_ISSUE_6950 = """
[nlp]
lang = "en"
pipeline = ["tok2vec", "tagger"]

[components]

[components.tok2vec]
factory = "tok2vec"

[components.tok2vec.model]
@architectures = "spacy.Tok2Vec.v1"

[components.tok2vec.model.embed]
@architectures = "spacy.MultiHashEmbed.v1"
width = ${components.tok2vec.model.encode:width}
attrs = ["NORM","PREFIX","SUFFIX","SHAPE"]
rows = [5000,2500,2500,2500]
include_static_vectors = false

[components.tok2vec.model.encode]
@architectures = "spacy.MaxoutWindowEncoder.v1"
width = 96
depth = 4
window_size = 1
maxout_pieces = 3

[components.ner]
factory = "ner"

[components.tagger]
factory = "tagger"

[components.tagger.model]
@architectures = "spacy.Tagger.v2"
nO = null

[components.tagger.model.tok2vec]
@architectures = "spacy.Tok2VecListener.v1"
width = ${components.tok2vec.model.encode:width}
upstream = "*"
"""


@pytest.mark.issue(6950)
def test_issue6950():
    """Test that the nlp object with initialized tok2vec with listeners pickles
    correctly (and doesn't have lambdas).
    """
    nlp = English.from_config(load_config_from_str(CONFIG_ISSUE_6950))
    nlp.initialize(lambda: [Example.from_dict(nlp.make_doc("hello"), {"tags": ["V"]})])
    pickle.dumps(nlp)
    nlp("hello")
    pickle.dumps(nlp)


def test_serialize_language_meta_disk(meta_data):
    language = Language(meta=meta_data)
    with make_tempdir() as d:
        language.to_disk(d)
        new_language = Language().from_disk(d)
    assert new_language.meta == language.meta


def test_serialize_with_custom_tokenizer():
    """Test that serialization with custom tokenizer works without token_match.
    See: https://support.prodi.gy/t/how-to-save-a-custom-tokenizer/661/2
    """
    prefix_re = re.compile(r"""1/|2/|:[0-9][0-9][A-K]:|:[0-9][0-9]:""")
    suffix_re = re.compile(r"""""")
    infix_re = re.compile(r"""[~]""")

    def custom_tokenizer(nlp):
        return Tokenizer(
            nlp.vocab,
            {},
            prefix_search=prefix_re.search,
            suffix_search=suffix_re.search,
            infix_finditer=infix_re.finditer,
        )

    nlp = Language()
    nlp.tokenizer = custom_tokenizer(nlp)
    with make_tempdir() as d:
        nlp.to_disk(d)


def test_serialize_language_exclude(meta_data):
    name = "name-in-fixture"
    nlp = Language(meta=meta_data)
    assert nlp.meta["name"] == name
    new_nlp = Language().from_bytes(nlp.to_bytes())
    assert new_nlp.meta["name"] == name
    new_nlp = Language().from_bytes(nlp.to_bytes(), exclude=["meta"])
    assert not new_nlp.meta["name"] == name
    new_nlp = Language().from_bytes(nlp.to_bytes(exclude=["meta"]))
    assert not new_nlp.meta["name"] == name



from libcpp.vector cimport vector

from ..structs cimport SpanC


cdef class SpanGroup:
    cdef public object _doc_ref
    cdef public str name
    cdef public dict attrs
    cdef vector[SpanC] c

    cdef void push_back(self, SpanC span) nogil



import pytest

from spacy.tokens import Doc

from ..util import apply_transition_sequence


def test_parser_space_attachment(en_vocab):
    # fmt: off
    words = ["This", "is", "a", "test", ".", "\n", "To", "ensure", " ", "spaces", "are", "attached", "well", "."]
    heads = [1, 1, 3, 1, 1, 4, 7, 11, 7, 11, 11, 11, 11, 11]
    # fmt: on
    deps = ["dep"] * len(heads)
    doc = Doc(en_vocab, words=words, heads=heads, deps=deps)
    for sent in doc.sents:
        if len(sent) == 1:
            assert not sent[-1].is_space


def test_parser_sentence_space(en_vocab):
    # fmt: off
    words = ["I", "look", "forward", "to", "using", "Thingamajig", ".", " ", "I", "'ve", "been", "told", "it", "will", "make", "my", "life", "easier", "..."]
    heads = [1, 1, 1, 1, 3, 4, 1, 6, 11, 11, 11, 11, 14, 14, 11, 16, 17, 14, 11]
    deps = ["nsubj", "ROOT", "advmod", "prep", "pcomp", "dobj", "punct", "",
            "nsubjpass", "aux", "auxpass", "ROOT", "nsubj", "aux", "ccomp",
            "poss", "nsubj", "ccomp", "punct"]
    # fmt: on
    doc = Doc(en_vocab, words=words, heads=heads, deps=deps)
    assert len(list(doc.sents)) == 2


@pytest.mark.skip(
    reason="The step_through API was removed (but should be brought back)"
)
def test_parser_space_attachment_leading(en_vocab, en_parser):
    words = ["\t", "\n", "This", "is", "a", "sentence", "."]
    heads = [1, 2, 2, 4, 2, 2]
    doc = Doc(en_vocab, words=words, heads=heads)
    assert doc[0].is_space
    assert doc[1].is_space
    assert doc[2].text == "This"
    with en_parser.step_through(doc) as stepwise:
        pass
    assert doc[0].head.i == 2
    assert doc[1].head.i == 2
    assert stepwise.stack == set([2])


@pytest.mark.skip(
    reason="The step_through API was removed (but should be brought back)"
)
def test_parser_space_attachment_intermediate_trailing(en_vocab, en_parser):
    words = ["This", "is", "\t", "a", "\t\n", "\n", "sentence", ".", "\n\n", "\n"]
    heads = [1, 1, 1, 5, 3, 1, 1, 6]
    transition = ["L-nsubj", "S", "L-det", "R-attr", "D", "R-punct"]
    doc = Doc(en_vocab, words=words, heads=heads)
    assert doc[2].is_space
    assert doc[4].is_space
    assert doc[5].is_space
    assert doc[8].is_space
    assert doc[9].is_space
    apply_transition_sequence(en_parser, doc, transition)
    for token in doc:
        assert token.dep != 0 or token.is_space
    assert [token.head.i for token in doc] == [1, 1, 1, 6, 3, 3, 1, 1, 7, 7]


@pytest.mark.parametrize("text,length", [(["\n"], 1), (["\n", "\t", "\n\n", "\t"], 4)])
@pytest.mark.skip(
    reason="The step_through API was removed (but should be brought back)"
)
def test_parser_space_attachment_space(en_parser, text, length):
    doc = Doc(en_parser.vocab, words=text)
    assert len(doc) == length
    with en_parser.step_through(doc) as _:  # noqa: F841
        pass
    assert doc[0].is_space
    for token in doc:
        assert token.head.i == length - 1



import itertools
import random
from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, Iterator, List, Optional, Tuple

from ..util import registry
from .example import Example
from .iob_utils import _doc_to_biluo_tags_with_partial, split_bilu_label

if TYPE_CHECKING:
    from ..language import Language  # noqa: F401


@registry.augmenters("spacy.combined_augmenter.v1")
def create_combined_augmenter(
    lower_level: float,
    orth_level: float,
    orth_variants: Optional[Dict[str, List[Dict]]],
    whitespace_level: float,
    whitespace_per_token: float,
    whitespace_variants: Optional[List[str]],
) -> Callable[["Language", Example], Iterator[Example]]:
    """Create a data augmentation callback that uses orth-variant replacement.
    The callback can be added to a corpus or other data iterator during training.

    lower_level (float): The percentage of texts that will be lowercased.
    orth_level (float): The percentage of texts that will be augmented.
    orth_variants (Optional[Dict[str, List[Dict]]]): A dictionary containing the
        single and paired orth variants. Typically loaded from a JSON file.
    whitespace_level (float): The percentage of texts that will have whitespace
        tokens inserted.
    whitespace_per_token (float): The number of whitespace tokens to insert in
        the modified doc as a percentage of the doc length.
    whitespace_variants (Optional[List[str]]): The whitespace token texts.
    RETURNS (Callable[[Language, Example], Iterator[Example]]): The augmenter.
    """
    return partial(
        combined_augmenter,
        lower_level=lower_level,
        orth_level=orth_level,
        orth_variants=orth_variants,
        whitespace_level=whitespace_level,
        whitespace_per_token=whitespace_per_token,
        whitespace_variants=whitespace_variants,
    )


def combined_augmenter(
    nlp: "Language",
    example: Example,
    *,
    lower_level: float = 0.0,
    orth_level: float = 0.0,
    orth_variants: Optional[Dict[str, List[Dict]]] = None,
    whitespace_level: float = 0.0,
    whitespace_per_token: float = 0.0,
    whitespace_variants: Optional[List[str]] = None,
) -> Iterator[Example]:
    if random.random() < lower_level:
        example = make_lowercase_variant(nlp, example)
    if orth_variants and random.random() < orth_level:
        raw_text = example.text
        orig_dict = example.to_dict()
        orig_dict["doc_annotation"]["entities"] = _doc_to_biluo_tags_with_partial(
            example.reference
        )
        variant_text, variant_token_annot = make_orth_variants(
            nlp,
            raw_text,
            orig_dict["token_annotation"],
            orth_variants,
            lower=False,
        )
        orig_dict["token_annotation"] = variant_token_annot
        example = example.from_dict(nlp.make_doc(variant_text), orig_dict)
    if whitespace_variants and random.random() < whitespace_level:
        for _ in range(int(len(example.reference) * whitespace_per_token)):
            example = make_whitespace_variant(
                nlp,
                example,
                random.choice(whitespace_variants),
                random.randrange(0, len(example.reference)),
            )
    yield example


@registry.augmenters("spacy.orth_variants.v1")
def create_orth_variants_augmenter(
    level: float, lower: float, orth_variants: Dict[str, List[Dict]]
) -> Callable[["Language", Example], Iterator[Example]]:
    """Create a data augmentation callback that uses orth-variant replacement.
    The callback can be added to a corpus or other data iterator during training.

    level (float): The percentage of texts that will be augmented.
    lower (float): The percentage of texts that will be lowercased.
    orth_variants (Dict[str, List[Dict]]): A dictionary containing
        the single and paired orth variants. Typically loaded from a JSON file.
    RETURNS (Callable[[Language, Example], Iterator[Example]]): The augmenter.
    """
    return partial(
        orth_variants_augmenter, orth_variants=orth_variants, level=level, lower=lower
    )


@registry.augmenters("spacy.lower_case.v1")
def create_lower_casing_augmenter(
    level: float,
) -> Callable[["Language", Example], Iterator[Example]]:
    """Create a data augmentation callback that converts documents to lowercase.
    The callback can be added to a corpus or other data iterator during training.

    level (float): The percentage of texts that will be augmented.
    RETURNS (Callable[[Language, Example], Iterator[Example]]): The augmenter.
    """
    return partial(lower_casing_augmenter, level=level)


def dont_augment(nlp: "Language", example: Example) -> Iterator[Example]:
    yield example


def lower_casing_augmenter(
    nlp: "Language", example: Example, *, level: float
) -> Iterator[Example]:
    if random.random() >= level:
        yield example
    else:
        yield make_lowercase_variant(nlp, example)


def make_lowercase_variant(nlp: "Language", example: Example):
    example_dict = example.to_dict()
    example_dict["doc_annotation"]["entities"] = _doc_to_biluo_tags_with_partial(
        example.reference
    )
    doc = nlp.make_doc(example.text.lower())
    example_dict["token_annotation"]["ORTH"] = [t.lower_ for t in example.reference]
    return example.from_dict(doc, example_dict)


def orth_variants_augmenter(
    nlp: "Language",
    example: Example,
    orth_variants: Dict[str, List[Dict]],
    *,
    level: float = 0.0,
    lower: float = 0.0,
) -> Iterator[Example]:
    if random.random() >= level:
        yield example
    else:
        raw_text = example.text
        orig_dict = example.to_dict()
        orig_dict["doc_annotation"]["entities"] = _doc_to_biluo_tags_with_partial(
            example.reference
        )
        variant_text, variant_token_annot = make_orth_variants(
            nlp,
            raw_text,
            orig_dict["token_annotation"],
            orth_variants,
            lower=raw_text is not None and random.random() < lower,
        )
        orig_dict["token_annotation"] = variant_token_annot
        yield example.from_dict(nlp.make_doc(variant_text), orig_dict)


def make_orth_variants(
    nlp: "Language",
    raw: str,
    token_dict: Dict[str, List[str]],
    orth_variants: Dict[str, List[Dict[str, List[str]]]],
    *,
    lower: bool = False,
) -> Tuple[str, Dict[str, List[str]]]:
    words = token_dict.get("ORTH", [])
    tags = token_dict.get("TAG", [])
    # keep unmodified if words are not defined
    if not words:
        return raw, token_dict
    if lower:
        words = [w.lower() for w in words]
        raw = raw.lower()
    # if no tags, only lowercase
    if not tags:
        token_dict["ORTH"] = words
        return raw, token_dict
    # single variants
    ndsv = orth_variants.get("single", [])
    punct_choices = [random.choice(x["variants"]) for x in ndsv]
    for word_idx in range(len(words)):
        for punct_idx in range(len(ndsv)):
            if (
                tags[word_idx] in ndsv[punct_idx]["tags"]
                and words[word_idx] in ndsv[punct_idx]["variants"]
            ):
                words[word_idx] = punct_choices[punct_idx]
    # paired variants
    ndpv = orth_variants.get("paired", [])
    punct_choices = [random.choice(x["variants"]) for x in ndpv]
    for word_idx in range(len(words)):
        for punct_idx in range(len(ndpv)):
            if tags[word_idx] in ndpv[punct_idx]["tags"] and words[
                word_idx
            ] in itertools.chain.from_iterable(ndpv[punct_idx]["variants"]):
                # backup option: random left vs. right from pair
                pair_idx = random.choice([0, 1])
                # best option: rely on paired POS tags like `` / ''
                if len(ndpv[punct_idx]["tags"]) == 2:
                    pair_idx = ndpv[punct_idx]["tags"].index(tags[word_idx])
                # next best option: rely on position in variants
                # (may not be unambiguous, so order of variants matters)
                else:
                    for pair in ndpv[punct_idx]["variants"]:
                        if words[word_idx] in pair:
                            pair_idx = pair.index(words[word_idx])
                words[word_idx] = punct_choices[punct_idx][pair_idx]
    token_dict["ORTH"] = words
    raw = construct_modified_raw_text(token_dict)
    return raw, token_dict


def make_whitespace_variant(
    nlp: "Language",
    example: Example,
    whitespace: str,
    position: int,
) -> Example:
    """Insert the whitespace token at the specified token offset in the doc.
    This is primarily intended for v2-compatible training data that doesn't
    include links or spans. If the document includes links, spans, or partial
    dependency annotation, it is returned without modifications.

    The augmentation follows the basics of the v2 space attachment policy, but
    without a distinction between "real" and other tokens, so space tokens
    may be attached to space tokens:
    - at the beginning of a sentence attach the space token to the following
      token
    - otherwise attach the space token to the preceding token

    The augmenter does not attempt to consolidate adjacent whitespace in the
    same way that the tokenizer would.

    The following annotation is used for the space token:
    TAG: "_SP"
    MORPH: ""
    POS: "SPACE"
    LEMMA: ORTH
    DEP: "dep"
    SENT_START: False

    The annotation for each attribute is only set for the space token if there
    is already at least partial annotation for that attribute in the original
    example.

    RETURNS (Example): Example with one additional space token.
    """
    example_dict = example.to_dict()
    example_dict["doc_annotation"]["entities"] = _doc_to_biluo_tags_with_partial(
        example.reference
    )
    doc_dict = example_dict.get("doc_annotation", {})
    token_dict = example_dict.get("token_annotation", {})
    # returned unmodified if:
    # - doc is empty
    # - words are not defined
    # - links are defined (only character-based offsets, which is more a quirk
    #   of Example.to_dict than a technical constraint)
    # - spans are defined
    # - there are partial dependencies
    if (
        len(example.reference) == 0
        or "ORTH" not in token_dict
        or len(doc_dict.get("links", [])) > 0
        or len(example.reference.spans) > 0
        or (
            example.reference.has_annotation("DEP")
            and not example.reference.has_annotation("DEP", require_complete=True)
        )
    ):
        return example
    words = token_dict.get("ORTH", [])
    length = len(words)
    assert 0 <= position <= length
    if example.reference.has_annotation("ENT_TYPE"):
        # I-ENTITY if between B/I-ENTITY and I/L-ENTITY otherwise O
        entity = "O"
        if position > 1 and position < length:
            ent_prev = doc_dict["entities"][position - 1]
            ent_next = doc_dict["entities"][position]
            if "-" in ent_prev and "-" in ent_next:
                ent_iob_prev, ent_type_prev = split_bilu_label(ent_prev)
                ent_iob_next, ent_type_next = split_bilu_label(ent_next)
                if (
                    ent_iob_prev in ("B", "I")
                    and ent_iob_next in ("I", "L")
                    and ent_type_prev == ent_type_next
                ):
                    entity = f"I-{ent_type_prev}"
        doc_dict["entities"].insert(position, entity)
    else:
        del doc_dict["entities"]
    token_dict["ORTH"].insert(position, whitespace)
    token_dict["SPACY"].insert(position, False)
    if example.reference.has_annotation("TAG"):
        token_dict["TAG"].insert(position, "_SP")
    else:
        del token_dict["TAG"]
    if example.reference.has_annotation("LEMMA"):
        token_dict["LEMMA"].insert(position, whitespace)
    else:
        del token_dict["LEMMA"]
    if example.reference.has_annotation("POS"):
        token_dict["POS"].insert(position, "SPACE")
    else:
        del token_dict["POS"]
    if example.reference.has_annotation("MORPH"):
        token_dict["MORPH"].insert(position, "")
    else:
        del token_dict["MORPH"]
    if example.reference.has_annotation("DEP", require_complete=True):
        if position == 0:
            token_dict["HEAD"].insert(position, 0)
        else:
            token_dict["HEAD"].insert(position, position - 1)
        for i in range(len(token_dict["HEAD"])):
            if token_dict["HEAD"][i] >= position:
                token_dict["HEAD"][i] += 1
        token_dict["DEP"].insert(position, "dep")
    else:
        del token_dict["HEAD"]
        del token_dict["DEP"]
    if example.reference.has_annotation("SENT_START"):
        token_dict["SENT_START"].insert(position, False)
    else:
        del token_dict["SENT_START"]
    raw = construct_modified_raw_text(token_dict)
    return Example.from_dict(nlp.make_doc(raw), example_dict)


def construct_modified_raw_text(token_dict):
    """Construct modified raw text from words and spaces."""
    raw = ""
    for orth, spacy in zip(token_dict["ORTH"], token_dict["SPACY"]):
        raw += orth
        if spacy:
            raw += " "
    return raw



from ...attrs import LIKE_NUM

_num_words = [
    "нуль",
    "ноль",
    "бер",
    "ике",
    "өч",
    "дүрт",
    "биш",
    "алты",
    "җиде",
    "сигез",
    "тугыз",
    "ун",
    "унбер",
    "унике",
    "унөч",
    "ундүрт",
    "унбиш",
    "уналты",
    "унҗиде",
    "унсигез",
    "унтугыз",
    "егерме",
    "утыз",
    "кырык",
    "илле",
    "алтмыш",
    "җитмеш",
    "сиксән",
    "туксан",
    "йөз",
    "мең",
    "төмән",
    "миллион",
    "миллиард",
    "триллион",
    "триллиард",
]


def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    if text in _num_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}



from ..char_classes import (
    ALPHA,
    ALPHA_LOWER,
    ALPHA_UPPER,
    CONCAT_QUOTES,
    LIST_ELLIPSES,
    LIST_ICONS,
)
from ..punctuation import TOKENIZER_SUFFIXES

_quotes = CONCAT_QUOTES.replace("'", "")

_infixes = (
    LIST_ELLIPSES
    + LIST_ICONS
    + [
        r"(?<=[{al}])\.(?=[{au}])".format(al=ALPHA_LOWER, au=ALPHA_UPPER),
        r"(?<=[{a}])[,!?](?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}])[<>=](?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}]):(?=[{a}])".format(a=ALPHA_UPPER),
        r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}])([{q}\)\]\(\[])(?=[{a}])".format(a=ALPHA, q=_quotes),
        r"(?<=[{a}])--(?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}0-9])[<>=/](?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}0-9]):(?=[{a}])".format(a=ALPHA_UPPER),
    ]
)

_suffixes = [
    suffix
    for suffix in TOKENIZER_SUFFIXES
    if suffix not in ["'s", "'S", "’s", "’S", r"\'"]
]
_suffixes += [r"(?<=[^sSxXzZ])\'"]


TOKENIZER_INFIXES = _infixes
TOKENIZER_SUFFIXES = _suffixes



from ...symbols import NORM, ORTH
from ...util import update_exc
from ..tokenizer_exceptions import BASE_EXCEPTIONS
from ._tokenizer_exceptions_list import MS_BASE_EXCEPTIONS

# Daftar singkatan dan Akronim dari:
# https://ms.wiktionary.org/wiki/Wiktionary:Senarai_akronim_dan_singkatan

_exc = {}

for orth in MS_BASE_EXCEPTIONS:
    _exc[orth] = [{ORTH: orth}]
    orth_title = orth.title()
    _exc[orth_title] = [{ORTH: orth_title}]
    orth_caps = orth.upper()
    _exc[orth_caps] = [{ORTH: orth_caps}]
    orth_lower = orth.lower()
    _exc[orth_lower] = [{ORTH: orth_lower}]
    orth_first_upper = orth[0].upper() + orth[1:]
    _exc[orth_first_upper] = [{ORTH: orth_first_upper}]
    if "-" in orth:
        orth_title = "-".join([part.title() for part in orth.split("-")])
        _exc[orth_title] = [{ORTH: orth_title}]
        orth_caps = "-".join([part.upper() for part in orth.split("-")])
        _exc[orth_caps] = [{ORTH: orth_caps}]

for exc_data in [
    {ORTH: "Jan.", NORM: "Januari"},
    {ORTH: "Feb.", NORM: "Februari"},
    {ORTH: "Mac.", NORM: "Mac"},
    {ORTH: "Apr.", NORM: "April"},
    {ORTH: "Jun.", NORM: "Jun"},
    {ORTH: "Jul.", NORM: "Julai"},
    {ORTH: "Ogos.", NORM: "Ogos"},
    {ORTH: "Sep.", NORM: "September"},
    {ORTH: "Okt.", NORM: "Oktober"},
    {ORTH: "Nov.", NORM: "November"},
    {ORTH: "Dis.", NORM: "Disember"},
]:
    _exc[exc_data[ORTH]] = [exc_data]

_other_exc = {
    "do'a": [{ORTH: "do'a", NORM: "doa"}],
    "jum'at": [{ORTH: "jum'at", NORM: "Jumat"}],
    "Jum'at": [{ORTH: "Jum'at", NORM: "Jumat"}],
    "la'nat": [{ORTH: "la'nat", NORM: "laknat"}],
    "ma'af": [{ORTH: "ma'af", NORM: "maaf"}],
    "mu'jizat": [{ORTH: "mu'jizat", NORM: "mukjizat"}],
    "Mu'jizat": [{ORTH: "Mu'jizat", NORM: "mukjizat"}],
    "ni'mat": [{ORTH: "ni'mat", NORM: "nikmat"}],
    "raka'at": [{ORTH: "raka'at", NORM: "rakaat"}],
    "ta'at": [{ORTH: "ta'at", NORM: "taat"}],
}

_exc.update(_other_exc)

for orth in [
    "1 Kor.",
    "1 Ptr.",
    "1 Raj.",
    "1 Sam.",
    "1 Taw.",
    "1 Tes.",
    "1 Tim.",
    "1 Yoh.",
    "1Ch.",
    "1Co.",
    "1Jo.",
    "1Ki.",
    "1Pe.",
    "1Sa.",
    "1Th.",
    "1Ti.",
    "2 Kor.",
    "2 Ptr.",
    "2 Raj.",
    "2 Sam.",
    "2 Taw.",
    "2 Tes.",
    "2 Tim.",
    "2 Yoh.",
    "2Ch.",
    "2Co.",
    "2Jo.",
    "2Ki.",
    "2Pe.",
    "2Sa.",
    "2Th.",
    "2Ti.",
    "3 Yoh.",
    "3D",
    "3F",
    "3Jo.",
    "3M",
    "8MP",
    "AA",
    "AAAAAA",
    "AB",
    "Abd.",
    "ABC",
    "ABIM",
    "ABM",
    "ABMI",
    "ABS",
    "AC",
    "Ac",
    "ACAPLPL",
    "Act.",
    "AD",
    "AD LIB",
    "ADAM",
    "ADB",
    "ADD",
    "ADIL",
    "ADN",
    "ADR",
    "ADRI",
    "ADSL",
    "ADUN",
    "AFAS",
    "AFTA",
    "Ag",
    "AGMARIS",
    "AH",
    "AI",
    "AIA",
    "AIDS",
    "AIJV",
    "AIM",
    "a/k",
    "ak",
    "AKN",
    "Al",
    "a/l",
    "AM",
    "Am",
    "Am.",
    "AMN",
    "Amo.",
    "AMPS",
    "Ams.",
    "AMWA",
    "AN",
    "a.n.",
    "ANGKASA",
    "ANM",
    "ANSI",
    "Ant.",
    "AOL",
    "AP",
    "a/p",
    "APD",
    "APEC",
    "API",
    "APIK",
    "APM",
    "APN",
    "APP",
    "Apr.",
    "APRI",
    "Ar",
    "Ar.",
    "ark.",
    "A.S.",
    "As",
    "a.s.",
    "ASA",
    "ASAS 50",
    "ASB",
    "ASCII",
    "ASEAN",
    "ASEAN+3",
    "ASEM",
    "a.s.f.",
    "ASN",
    "a.s.o.",
    "ASP",
    "Ast.",
    "A.T.",
    "At",
    "ATM",
    "a.t.r.",
    "ATUR",
    "Au",
    "AURI",
    "Aug.",
    "AWOL",
    "Ayb.",
    "B",
    "BA",
    "Ba",
    "BAC",
    "BAFIA",
    "BAM",
    "BANANA",
    "BAPP",
    "BASF",
    "BATA",
    "BB",
    "BBC",
    "BBE",
    "BBS",
    "BC",
    "BCG",
    "BCIC",
    "b.d.",
    "BDSSHAM",
    "Be",
    "BEER",
    "BERNAMA",
    "Bh",
    "b.h.",
    "Bhd.",
    "Bi",
    "BIDS",
    "Bil.",
    "bil.",
    "BIMP-EAGA",
    "Bio.",
    "BIOS",
    "BITMB",
    "BJ",
    "Bk",
    "b.k.",
    "BKAL",
    "bkn.",
    "BKP",
    "BL",
    "BLR",
    "BM",
    "BMI",
    "BMW",
    "BN",
    "BNM",
    "BO",
    "BOJ",
    "BOO",
    "BOP",
    "BOT",
    "BP",
    "b.p.",
    "BPA",
    "BPAs",
    "bpd.",
    "BPIMB",
    "BPM",
    "BPO",
    "BPPH",
    "Br",
    "Br.",
    "BSA",
    "B.Sc.",
    "B.Sh.",
    "b.s.j.",
    "BSN",
    "Bt.",
    "bt.",
    "BWT",
    "BYOB",
    "C",
    "C.",
    "C/E",
    "Ca",
    "CAAM",
    "CAD",
    "CAM",
    "CATV",
    "CBS",
    "CBT",
    "CC",
    "CCD",
    "CCM",
    "CCR",
    "cct-km",
    "CCTV",
    "CCU",
    "CD",
    "Cd",
    "CD-ROM",
    "CD-RW",
    "CDRC",
    "Ce",
    "CEO",
    "CEPT",
    "Cetak",
    "Cf",
    "CFO",
    "CFTC",
    "CGC",
    "CGI",
    "CH",
    "CIA",
    "CIAST",
    "CID",
    "CIDB",
    "CIQ",
    "CKD",
    "CL",
    "Cl",
    "c.l.",
    "CLI",
    "CLOB",
    "CM",
    "Cm",
    "cm.",
    "CMAG",
    "CMI",
    "CMP",
    "CNN",
    "Co",
    "COD",
    "Col.",
    "COLA",
    "COMDEX",
    "CP",
    "CPI",
    "CPO",
    "CPR",
    "CPU",
    "Cr",
    "CRDF",
    "Cs",
    "CST",
    "CT",
    "CTIP",
    "CTRM",
    "Cu",
    "CUEPACS",
    "D-8",
    "d/a",
    "DAGS",
    "Dan.",
    "DANCED",
    "DAP",
    "DARA",
    "Db",
    "DBKL",
    "DBP",
    "DBR",
    "DC",
    "DDA",
    "DDT",
    "DEB",
    "Dec.",
    "Deu.",
    "DFIs",
    "dgn.",
    "DHL",
    "DIBML",
    "DIN",
    "Dis.",
    "DJ",
    "d.l.l.",
    "dlm.",
    "dng.",
    "DNS",
    "DO",
    "DOA",
    "DOE",
    "DOF",
    "DOSH",
    "doz.",
    "DPPS",
    "Dr.",
    "dr.",
    "drp.",
    "drpd.",
    "Ds",
    "d.sb.",
    "d.st.",
    "DSTN2",
    "Dt.",
    "DTAs",
    "DTMF",
    "DTP",
    "DTV",
    "DUBES",
    "DUNHILL",
    "DV8",
    "DVD",
    "DVE",
    "DVS",
    "dw.t.",
    "Dy",
    "DYMM",
    "E",
    "E-Commerce",
    "E-Dagang",
    "E&E",
    "E-Faraid",
    "E-Government",
    "E-Kerajaan",
    "E-Mail",
    "E-Services",
    "E-Village",
    "E-Zine",
    "EALAF",
    "EBI",
    "EBP",
    "EC",
    "ECAFE",
    "Ecc.",
    "ECI",
    "ECM",
    "ECOSOC",
    "ECP",
    "ECR",
    "EDI",
    "EE",
    "EEC",
    "Ef.",
    "EG",
    "Eko.",
    "EKS",
    "ELWS",
    "ELX",
    "EMI",
    "EMUs",
    "En.",
    "EP",
    "EPF",
    "Eph.",
    "EPP",
    "EPS",
    "EPU",
    "ER",
    "Er",
    "ERL",
    "ERT",
    "Es",
    "ESCAP",
    "ESOS",
    "ESP",
    "EST",
    "Est.",
    "ET",
    "ETA",
    "ETACS",
    "ETC",
    "ETD",
    "EU",
    "Eu",
    "EVIAN",
    "Exim Bank",
    "Exo.",
    "Eze.",
    "Ezr.",
    "F",
    "FAM",
    "FAMA",
    "FAO",
    "FAQ",
    "FAX",
    "FBI",
    "FC",
    "FCA",
    "FCC",
    "FDI",
    "FE",
    "Fe",
    "f.e.",
    "Feb.",
    "FELCRA",
    "FELDA",
    "FI",
    "FIA 1993",
    "FIAT",
    "FIC",
    "FIDA",
    "FIFA",
    "FIMA",
    "Fiz.",
    "Flm.",
    "Flp.",
    "FM",
    "Fm",
    "FMUTM",
    "FO",
    "FOA",
    "FOB",
    "FOC",
    "FOMCA",
    "FORD",
    "Fr",
    "FRIM",
    "FRTI",
    "FSMP",
    "FTA",
    "FTE",
    "FTP",
    "G",
    "g.",
    "G15",
    "G77",
    "Ga",
    "GAC",
    "GACM",
    "Gal.",
    "GAPENA",
    "GATS",
    "GATT",
    "GB",
    "Gbps.",
    "Gd",
    "GDP",
    "Ge",
    "GEC",
    "Gen.",
    "Geo.",
    "Geog.",
    "Gerakan",
    "GH",
    "GIF",
    "GII",
    "GIS",
    "GITIC",
    "GITN",
    "GJ",
    "GLCs",
    "GM",
    "GMBH",
    "GMI",
    "GMT",
    "GNP",
    "GNS",
    "GOLD",
    "GP",
    "GPC",
    "GPIM",
    "GPMS",
    "GPO",
    "GPP",
    "GPS",
    "GRO",
    "GRS",
    "GSMC",
    "GST",
    "GTZ",
    "GUI",
    "GWh.",
    "H",
    "Ha",
    "Hab.",
    "Hag.",
    "Hak.",
    "ham",
    "hb.",
    "HCI",
    "HDTV",
    "He",
    "Heb.",
    "Hf",
    "Hg",
    "HI-FI",
    "HIS",
    "HIV",
    "Hj.",
    "HMS",
    "Ho",
    "Hos.",
    "HP",
    "HRDC",
    "HRDF",
    "HRMIS",
    "Hs",
    "Hut.",
    "I",
    "I/O",
    "IA",
    "IAA",
    "IADPs",
    "IB",
    "i.b.",
    "IBA",
    "IBFIM",
    "IBG",
    "Ibr.",
    "IBRD",
    "IBS",
    "IC",
    "ICA",
    "ICBM",
    "ICFM",
    "ICI",
    "ICM",
    "ICOR",
    "ICP",
    "ICT",
    "ICU",
    "ID",
    "Id.",
    "IDB",
    "IDFR",
    "IE",
    "i.e.",
    "IFSB",
    "IGAs",
    "IGS",
    "IHP",
    "IHPG",
    "IIM",
    "IINA",
    "IKKL",
    "IKP",
    "IKPH",
    "IKS",
    "Im.",
    "IMD",
    "IMF",
    "IMP2",
    "IMR",
    "IMS-GT",
    "IMT-GT",
    "In",
    "in.",
    "INFRA",
    "INSEP",
    "INSPEN",
    "INTAN",
    "IOFC",
    "IOU",
    "IP",
    "IPA",
    "IPBA",
    "IPCs",
    "IPEBP",
    "IPI",
    "IPKIM",
    "IPKPM",
    "IPO",
    "IPP",
    "IPPM",
    "IPPPM",
    "i.pt.",
    "IPTAR",
    "IPTNM",
    "IQR",
    "Ir",
    "IRA",
    "IRPA",
    "IRS",
    "i.s.",
    "ISA",
    "Isa.",
    "ISDN",
    "ISMM",
    "ISO",
    "ISP",
    "ist.",
    "IT",
    "i.t.",
    "ITA",
    "ITAF",
    "ITEX",
    "ITK",
    "ITM",
    "ITO",
    "ITRCo",
    "ITTA",
    "ITU",
    "JAK",
    "JAKIM",
    "Jam.",
    "Jan.",
    "Jb.",
    "JBIC",
    "JD",
    "JDA",
    "Jdg.",
    "Jer.",
    "Jh.",
    "JICA",
    "JJ",
    "Jk.",
    "JKKK",
    "jkps.",
    "JKR",
    "JMTI",
    "JOA",
    "Joe.",
    "Joh.",
    "Jon.",
    "Jos.",
    "JP",
    "JPA",
    "JPEG",
    "JPH",
    "JPJ",
    "JPSHK",
    "JPS",
    "JPT",
    "JRDA",
    "JSM",
    "JT",
    "Jud.",
    "Jul.",
    "Jun.",
    "JVC",
    "Jw.",
    "K",
    "K-Economy",
    "KADA",
    "KBE",
    "KBIA",
    "KBPA",
    "KBSM",
    "KD",
    "Kd.",
    "KDI",
    "KDN",
    "KDNK",
    "KE",
    "KEAP",
    "Kej.",
    "Kel.",
    "KEM",
    "KEMLU",
    "kep.",
    "Kg.",
    "kg.",
    "KGB",
    "KGK",
    "KH",
    "ki.",
    "Kid.",
    "KIK",
    "KIKMTT",
    "KIM",
    "Kim.",
    "Kis.",
    "KIX",
    "KKGSK",
    "KKK",
    "KKPPA",
    "KL",
    "Kl.",
    "KLCI",
    "KLIA",
    "KLIBOR",
    "KLIM",
    "KLM",
    "KLSE",
    "KM",
    "KMM",
    "KNK",
    "KO",
    "Kol.",
    "Kom.",
    "Komp.",
    "KOMSAS",
    "KPAI",
    "KPB",
    "KPBA",
    "KPC",
    "kpd.",
    "KPE",
    "KPIs",
    "KPPL",
    "KPPMS",
    "KPWM",
    "Kr",
    "KRM",
    "KSTI",
    "KT",
    "KTA",
    "KTABKL",
    "KTM",
    "KTMB",
    "kV",
    "kW",
    "kWh",
    "kWj",
    "KWSP",
    "LA",
    "La",
    "LABOR",
    "Lam.",
    "LAN",
    "LAPD",
    "LASER",
    "LAX",
    "lb.",
    "LC",
    "LCD",
    "LCHRF",
    "LCLY",
    "LED",
    "Lev.",
    "LFPR",
    "LFS",
    "LFX",
    "LGM",
    "Li",
    "LID",
    "Lin.",
    "LKN",
    "LKPM",
    "LKPP",
    "LKTP",
    "LKWJ",
    "LLB",
    "LLC",
    "LLN",
    "LLS",
    "LMSM",
    "LNG",
    "LOA",
    "LOBATA",
    "LOFSA",
    "LPG",
    "LPIP",
    "LPKI",
    "LPKLPL",
    "LPKN",
    "LPN",
    "LPP",
    "LPPK",
    "LPPM",
    "LPPP",
    "LPPTP",
    "Lr",
    "LRs",
    "LRT",
    "LS",
    "LTAKL",
    "LTD",
    "LTK",
    "Lu",
    "LUAS",
    "Luk.",
    "lw.",
    "lwn.",
    "M\n",
    "m",
    "M&A",
    "MAB",
    "MACRES",
    "MAD",
    "MADA",
    "MAGERAN",
    "MAHA",
    "MAHSURI",
    "Mal.",
    "MALINDO",
    "MAMPU",
    "Mar.",
    "MARA",
    "MARC",
    "MARDI",
    "MARLBORO",
    "MAS",
    "MASSA",
    "MASSCORP",
    "Mat.",
    "MATRADE",
    "MAVCAP",
    "MB",
    "MBA",
    "MBBS",
    "MBM",
    "MBO",
    "MBS",
    "MBTU",
    "MC",
    "MCA",
    "MCB",
    "MCSL",
    "MCSv5",
    "MD",
    "Md",
    "MDB",
    "MDC",
    "MDG",
    "MDV",
    "MEASAT",
    "MEATJ",
    "MECIB",
    "MEMO",
    "MENLU",
    "MEPS",
    "MES",
    "MESDAQ",
    "METEOR",
    "MFI",
    "MFIs",
    "MG",
    "Mg",
    "MGM",
    "MGR",
    "MGS",
    "MHA",
    "Mi.",
    "MIA",
    "MIB",
    "MIC",
    "Mic.",
    "MICE",
    "MIDA",
    "MIDF",
    "MIDI",
    "MIG",
    "MIGHT",
    "MII",
    "MIMOS",
    "MINDEF",
    "MINT",
    "mis.",
    "MIT",
    "MITC",
    "MITI",
    "Ml.",
    "MLNG",
    "mlpd.",
    "MM",
    "mm",
    "MMN",
    "mmscfd.",
    "MMU",
    "MMX",
    "Mn",
    "Mn.",
    "MNA",
    "MNCs",
    "MO",
    "Mo",
    "MOA",
    "MOD",
    "MODEM",
    "MOE",
    "MOH",
    "MOSTE",
    "MOSTI",
    "MOU",
    "MP",
    "MPB",
    "MPEG",
    "MPOB",
    "MPP",
    "mppa.",
    "MPPJ",
    "MPS",
    "MPTM",
    "MR",
    "m.r.",
    "MRB",
    "MRELB",
    "Mrk.",
    "MRRDB",
    "MS",
    "MS-DOS",
    "MSC",
    "MSG",
    "MSM",
    "Mt",
    "MTC",
    "MTCP",
    "MTD",
    "MTDC",
    "MTPB",
    "MTV",
    "Muz.",
    "MV",
    "MW",
    "MY",
    "MyKe",
    "Mzm.",
    "N",
    "N/A",
    "Na",
    "NAB",
    "NACIWID",
    "Nah.",
    "NAP",
    "NASA",
    "NATO",
    "NAV",
    "NB",
    "Nb",
    "NBA",
    "NBC",
    "NCR",
    "Nd",
    "NDP",
    "Ne",
    "NEAC",
    "NEC",
    "NEF",
    "Neh.",
    "NEP",
    "NEqO",
    "NERP",
    "NF",
    "NFPEs",
    "NG",
    "NGOs",
    "NGV",
    "NHEF",
    "NHHES",
    "NHK",
    "Ni",
    "NIDC",
    "NIH",
    "NIP",
    "NIPA",
    "NIS",
    "NISIR",
    "NITA",
    "NITC",
    "NITP",
    "NIV",
    "NLAC",
    "NMPBSP",
    "NMU",
    "No",
    "No.",
    "no.",
    "NOSS",
    "Nov.",
    "Np",
    "NPC",
    "NPCS",
    "NPL",
    "NRCC",
    "NRW",
    "NS",
    "Ns",
    "NSB",
    "NTA",
    "NTHRDC",
    "NTMP",
    "NTSC",
    "Num.",
    "NUTF",
    "NVP",
    "NVTC",
    "NWRC",
    "O",
    "Ob.",
    "Oba.",
    "OC",
    "OCPD",
    "Oct.",
    "OD",
    "ODA",
    "OECD",
    "OEM",
    "Ogo.",
    "OHQs",
    "OIC",
    "Okt.",
    "OPEC",
    "OPP",
    "OPP3",
    "OPR",
    "OS",
    "Os",
    "OSA",
    "OT",
    "OUG",
    "oz.",
    "P",
    "P&P",
    "PA",
    "Pa",
    "PABK",
    "PABX",
    "PAK",
    "PAKSI",
    "PAL",
    "PALL MALL",
    "PAS",
    "PATA",
    "PAWS",
    "Pb",
    "PBA",
    "PBB",
    "PBM",
    "PBP",
    "PBSM",
    "PBT",
    "PC",
    "PC(s)",
    "PCB",
    "PCIRITA",
    "PCM",
    "PCMCIA",
    "PCN",
    "PD",
    "Pd",
    "pd.",
    "PDS",
    "PE",
    "PEKEMAS",
    "PEMADAM",
    "PENA",
    "PENIS",
    "PERDANA",
    "PERKESO",
    "PERKIM",
    "PERNAS",
    "PERTAMA",
    "PERTIWI",
    "PESAKA",
    "PETA",
    "PETRONAS",
    "PGU",
    "Ph.",
    "PHD",
    "Phi.",
    "Phm.",
    "PIK",
    "PIKOM",
    "PIN",
    "PINTAS",
    "PIPM",
    "PISK",
    "PITA",
    "PIXEL",
    "PJ",
    "PJK",
    "PJKB",
    "PJP",
    "PKBM",
    "PKBTA",
    "PKEN",
    "Pkh.",
    "PKKM",
    "PKLPA",
    "PKM",
    "PKNS",
    "PKPIM",
    "PKPM",
    "PKR",
    "PKS",
    "Pl.",
    "p.l.",
    "PLA",
    "PLC",
    "PLCHP",
    "PLCs",
    "PLI",
    "PLT",
    "PLUS",
    "PLWS",
    "PM",
    "Pm",
    "PMM",
    "PMP",
    "PMR",
    "PMS",
    "Pn.",
    "PNAT",
    "PNS",
    "PO",
    "Po",
    "POCPA",
    "POKEMON",
    "Pol.",
    "POP",
    "PORIM",
    "PORLA",
    "PORTAFOAM",
    "PP",
    "PPA",
    "PPBE",
    "PPBK",
    "ppd.",
    "PPGM",
    "PPI",
    "PPK",
    "PPL",
    "PPM",
    "PPP",
    "PPPB",
    "PPPLM",
    "PPPM",
    "PPR",
    "PPRT",
    "PPS",
    "PPTM",
    "PPU",
    "PR",
    "Pr",
    "Pr.",
    "prb.",
    "PRI",
    "PRO",
    "Pro.",
    "Prof.",
    "PROSPER",
    "PROSTAR",
    "PROTON",
    "PS",
    "PSA",
    "Psa.",
    "PSCs",
    "PSDC",
    "PSDH",
    "Psi.",
    "PSKE",
    "PSRM",
    "PST",
    "PT",
    "Pt",
    "PTD",
    "PTP",
    "Pu",
    "PUNB",
    "QA",
    "QC",
    "QCC",
    "R&D",
    "RA",
    "Ra",
    "RAM",
    "RAPP",
    "Rat.",
    "Rb",
    "RCA",
    "RDA",
    "RDAs",
    "RDCs",
    "RE",
    "Re",
    "REHDA",
    "Rev.",
    "Rf",
    "Rg",
    "RGB",
    "Rh",
    "RI",
    "RIDA",
    "RIP",
    "RISDA",
    "r.l.",
    "RM",
    "Rm.",
    "RMKe-8",
    "Rn",
    "ROC",
    "ROM",
    "Rom.",
    "RPG",
    "RPS",
    "RRI",
    "RRIM",
    "RRJP",
    "RRP",
    "RSGC",
    "RSS",
    "RSVP",
    "Rt.",
    "RTA",
    "RTM",
    "Ru",
    "Rut.",
    "RWCR",
    "RX",
    "S",
    "S/N",
    "S&T",
    "S-VHS",
    "SA",
    "SAC",
    "SADCs",
    "SAGA",
    "SALCRA",
    "SALM",
    "SALT",
    "SAM",
    "SAP",
    "SARS",
    "Sas.",
    "s.a.w.",
    "SB",
    "Sb",
    "Sb.",
    "SBA",
    "SBB",
    "sbg.",
    "SBK",
    "SC",
    "Sc",
    "SCA",
    "SCADA",
    "SCANS",
    "SCSI",
    "SCuM",
    "SDCs",
    "Sdn. Bhd.",
    "sdr.",
    "SDRC",
    "Se",
    "SEATO",
    "SEB",
    "SECAM",
    "SEDCs",
    "SEFF",
    "Sej.",
    "SEMS",
    "Sep.",
    "Sept.",
    "SESB",
    "SESCo",
    "s.f.",
    "Sg",
    "SGPCA",
    "SGPPI",
    "SGPPKRM",
    "SGX",
    "Si",
    "Si.",
    "SIA 1983",
    "SIC",
    "SIM",
    "SING",
    "SIRIM",
    "SITTDEC",
    "sj.",
    "SKDTP",
    "SKM",
    "SKSM",
    "SL",
    "Sl.",
    "sl.",
    "SLMCH",
    "SLR",
    "SM",
    "Sm",
    "SMART",
    "SMEs",
    "SMEt",
    "SMIs",
    "SMIDEC",
    "SMIDP",
    "SMJK",
    "SMR",
    "SMS",
    "SMT",
    "SMTP",
    "SN",
    "Sn",
    "SOB",
    "SOCSO",
    "SOHO",
    "Son.",
    "SOS",
    "Sos.",
    "SP",
    "SPA",
    "SPAM",
    "SPCA",
    "SPKR",
    "SPLAM",
    "SPM",
    "SPNB",
    "SPSP",
    "t.",
    "Ta",
    "Tadb.",
    "TAF",
    "TAF-W",
    "Tani",
    "TAP",
    "TAR",
    "TARBI",
    "TB",
    "Tb",
    "TBA",
    "TBTP",
    "Tc",
    "TCPD",
    "TDCs",
    "Te",
    "TEKUN",
    "TELCO",
    "TELEX",
    "TEUs",
    "TFP",
    "TGV",
    "TH",
    "Th",
    "THIS",
    "Ti",
    "TICAD",
    "Tit.",
    "TKA",
    "Tks.",
    "Tl",
    "TLDM",
    "TM",
    "Tm",
    "TMB",
    "TMK",
    "TNB",
    "TNSB",
    "TNT",
    "TOEFL",
    "TP",
    "TPIM",
    "TPK",
    "TPPP",
    "TPPT",
    "TPSM",
    "TPUB",
    "TQM",
    "Tr.",
    "TRIPs",
    "tsb.",
    "tscf.",
    "t.sh.",
    "t.s.t.",
    "TT",
    "t.t.",
    "TUDM",
    "TV",
    "TVSMR",
    "TWAIN",
    "TX",
    "TYPHIrapid",
    "U",
    "Ubat",
    "UDA",
    "Udg.",
    "UFO",
    "UH",
    "UIA",
    "UiTM",
    "UK",
    "UKM",
    "UL",
    "Ul.",
    "ULC",
    "UM",
    "UMNO",
    "UMS",
    "UN",
    "UN/OSCAL",
    "UNCLE",
    "UNCTAD",
    "UNDP",
    "UNESCO",
    "UNFCCC",
    "UNFPA",
    "UNHCR",
    "UNICEF",
    "UNIMAS",
    "UNTAET",
    "UPE",
    "UPM",
    "UPS",
    "UPSR",
    "URL",
    "US",
    "USAINS",
    "USD",
    "USM",
    "USNO",
    "USS",
    "USSR",
    "UTC",
    "UTF",
    "utk.",
    "UTM",
    "V",
    "VAT",
    "VCC",
    "VCD",
    "VCR",
    "VD",
    "VDSC",
    "VGA",
    "VHF",
    "VHS",
    "VIP",
    "VMS",
    "VO",
    "VOA",
    "VoIP",
    "VR",
    "VSOP",
    "VW",
    "W",
    "W/O",
    "WAP",
    "WAY",
    "WC",
    "WDDM",
    "WDM",
    "WHO",
    "Why.",
    "WIM",
    "WPG",
    "WTO",
    "WWF",
    "WWW",
    "WYSIWYG",
    "Xe",
    "XO",
    "XXL",
    "Y",
    "Y2K",
    "YAB",
    "Yak.",
    "YAM",
    "YAS",
    "YB",
    "Yb",
    "Yeh.",
    "Yer.",
    "Yes.",
    "yg.",
    "Yl.",
    "YM",
    "YMCA",
    "Yoh.",
    "Yos.",
    "Y.Th.",
    "YTM",
    "Yud.",
    "Yun.",
    "Za.",
    "Zec.",
    "Zef.",
    "Zep.",
    "ZIP",
    "Zn",
    "Zr",
]:
    _exc[orth] = [{ORTH: orth}]

TOKENIZER_EXCEPTIONS = update_exc(BASE_EXCEPTIONS, _exc)



import warnings

import pytest
import srsly
from mock import Mock

from spacy.lang.en import English
from spacy.matcher import Matcher, PhraseMatcher
from spacy.tokens import Doc, Span
from spacy.vocab import Vocab

from ..util import make_tempdir


@pytest.mark.issue(3248)
def test_issue3248_1():
    """Test that the PhraseMatcher correctly reports its number of rules, not
    total number of patterns."""
    nlp = English()
    matcher = PhraseMatcher(nlp.vocab)
    matcher.add("TEST1", [nlp("a"), nlp("b"), nlp("c")])
    matcher.add("TEST2", [nlp("d")])
    assert len(matcher) == 2


@pytest.mark.issue(3331)
def test_issue3331(en_vocab):
    """Test that duplicate patterns for different rules result in multiple
    matches, one per rule.
    """
    matcher = PhraseMatcher(en_vocab)
    matcher.add("A", [Doc(en_vocab, words=["Barack", "Obama"])])
    matcher.add("B", [Doc(en_vocab, words=["Barack", "Obama"])])
    doc = Doc(en_vocab, words=["Barack", "Obama", "lifts", "America"])
    matches = matcher(doc)
    assert len(matches) == 2
    match_ids = [en_vocab.strings[matches[0][0]], en_vocab.strings[matches[1][0]]]
    assert sorted(match_ids) == ["A", "B"]


@pytest.mark.issue(3972)
def test_issue3972(en_vocab):
    """Test that the PhraseMatcher returns duplicates for duplicate match IDs."""
    matcher = PhraseMatcher(en_vocab)
    matcher.add("A", [Doc(en_vocab, words=["New", "York"])])
    matcher.add("B", [Doc(en_vocab, words=["New", "York"])])
    doc = Doc(en_vocab, words=["I", "live", "in", "New", "York"])
    matches = matcher(doc)

    assert len(matches) == 2

    # We should have a match for each of the two rules
    found_ids = [en_vocab.strings[ent_id] for (ent_id, _, _) in matches]
    assert "A" in found_ids
    assert "B" in found_ids


@pytest.mark.issue(4002)
def test_issue4002(en_vocab):
    """Test that the PhraseMatcher can match on overwritten NORM attributes."""
    matcher = PhraseMatcher(en_vocab, attr="NORM")
    pattern1 = Doc(en_vocab, words=["c", "d"])
    assert [t.norm_ for t in pattern1] == ["c", "d"]
    matcher.add("TEST", [pattern1])
    doc = Doc(en_vocab, words=["a", "b", "c", "d"])
    assert [t.norm_ for t in doc] == ["a", "b", "c", "d"]
    matches = matcher(doc)
    assert len(matches) == 1
    matcher = PhraseMatcher(en_vocab, attr="NORM")
    pattern2 = Doc(en_vocab, words=["1", "2"])
    pattern2[0].norm_ = "c"
    pattern2[1].norm_ = "d"
    assert [t.norm_ for t in pattern2] == ["c", "d"]
    matcher.add("TEST", [pattern2])
    matches = matcher(doc)
    assert len(matches) == 1


@pytest.mark.issue(4373)
def test_issue4373():
    """Test that PhraseMatcher.vocab can be accessed (like Matcher.vocab)."""
    matcher = Matcher(Vocab())
    assert isinstance(matcher.vocab, Vocab)
    matcher = PhraseMatcher(Vocab())
    assert isinstance(matcher.vocab, Vocab)


@pytest.mark.issue(4651)
def test_issue4651_with_phrase_matcher_attr():
    """Test that the EntityRuler PhraseMatcher is deserialized correctly using
    the method from_disk when the EntityRuler argument phrase_matcher_attr is
    specified.
    """
    text = "Spacy is a python library for nlp"
    nlp = English()
    patterns = [{"label": "PYTHON_LIB", "pattern": "spacy", "id": "spaCy"}]
    ruler = nlp.add_pipe("entity_ruler", config={"phrase_matcher_attr": "LOWER"})
    ruler.add_patterns(patterns)
    doc = nlp(text)
    res = [(ent.text, ent.label_, ent.ent_id_) for ent in doc.ents]
    nlp_reloaded = English()
    with make_tempdir() as d:
        file_path = d / "entityruler"
        ruler.to_disk(file_path)
        nlp_reloaded.add_pipe("entity_ruler").from_disk(file_path)
    doc_reloaded = nlp_reloaded(text)
    res_reloaded = [(ent.text, ent.label_, ent.ent_id_) for ent in doc_reloaded.ents]
    assert res == res_reloaded


@pytest.mark.issue(6839)
def test_issue6839(en_vocab):
    """Ensure that PhraseMatcher accepts Span as input"""
    # fmt: off
    words = ["I", "like", "Spans", "and", "Docs", "in", "my", "input", ",", "and", "nothing", "else", "."]
    # fmt: on
    doc = Doc(en_vocab, words=words)
    span = doc[:8]
    pattern = Doc(en_vocab, words=["Spans", "and", "Docs"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("SPACY", [pattern])
    matches = matcher(span)
    assert matches


@pytest.mark.issue(10643)
def test_issue10643(en_vocab):
    """Ensure overlapping terms can be removed from PhraseMatcher"""

    # fmt: off
    words = ["Only", "save", "out", "the", "binary", "data", "for", "the", "individual", "components", "."]
    # fmt: on
    doc = Doc(en_vocab, words=words)
    terms = {
        "0": Doc(en_vocab, words=["binary"]),
        "1": Doc(en_vocab, words=["binary", "data"]),
    }
    matcher = PhraseMatcher(en_vocab)
    for match_id, term in terms.items():
        matcher.add(match_id, [term])

    matches = matcher(doc)
    assert matches == [(en_vocab.strings["0"], 4, 5), (en_vocab.strings["1"], 4, 6)]

    matcher.remove("0")
    assert len(matcher) == 1
    new_matches = matcher(doc)
    assert new_matches == [(en_vocab.strings["1"], 4, 6)]

    matcher.remove("1")
    assert len(matcher) == 0
    no_matches = matcher(doc)
    assert not no_matches


def test_matcher_phrase_matcher(en_vocab):
    doc = Doc(en_vocab, words=["I", "like", "Google", "Now", "best"])
    # intermediate phrase
    pattern = Doc(en_vocab, words=["Google", "Now"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("COMPANY", [pattern])
    assert len(matcher(doc)) == 1
    # initial token
    pattern = Doc(en_vocab, words=["I"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("I", [pattern])
    assert len(matcher(doc)) == 1
    # initial phrase
    pattern = Doc(en_vocab, words=["I", "like"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("ILIKE", [pattern])
    assert len(matcher(doc)) == 1
    # final token
    pattern = Doc(en_vocab, words=["best"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("BEST", [pattern])
    assert len(matcher(doc)) == 1
    # final phrase
    pattern = Doc(en_vocab, words=["Now", "best"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("NOWBEST", [pattern])
    assert len(matcher(doc)) == 1


def test_phrase_matcher_length(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    assert len(matcher) == 0
    matcher.add("TEST", [Doc(en_vocab, words=["test"])])
    assert len(matcher) == 1
    matcher.add("TEST2", [Doc(en_vocab, words=["test2"])])
    assert len(matcher) == 2


def test_phrase_matcher_contains(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    matcher.add("TEST", [Doc(en_vocab, words=["test"])])
    assert "TEST" in matcher
    assert "TEST2" not in matcher


def test_phrase_matcher_add_new_api(en_vocab):
    doc = Doc(en_vocab, words=["a", "b"])
    patterns = [Doc(en_vocab, words=["a"]), Doc(en_vocab, words=["a", "b"])]
    matcher = PhraseMatcher(en_vocab)
    matcher.add("OLD_API", None, *patterns)
    assert len(matcher(doc)) == 2
    matcher = PhraseMatcher(en_vocab)
    on_match = Mock()
    matcher.add("OLD_API_CALLBACK", on_match, *patterns)
    assert len(matcher(doc)) == 2
    assert on_match.call_count == 2
    # New API: add(key: str, patterns: List[List[dict]], on_match: Callable)
    matcher = PhraseMatcher(en_vocab)
    matcher.add("NEW_API", patterns)
    assert len(matcher(doc)) == 2
    matcher = PhraseMatcher(en_vocab)
    on_match = Mock()
    matcher.add("NEW_API_CALLBACK", patterns, on_match=on_match)
    assert len(matcher(doc)) == 2
    assert on_match.call_count == 2


def test_phrase_matcher_repeated_add(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    # match ID only gets added once
    matcher.add("TEST", [Doc(en_vocab, words=["like"])])
    matcher.add("TEST", [Doc(en_vocab, words=["like"])])
    matcher.add("TEST", [Doc(en_vocab, words=["like"])])
    matcher.add("TEST", [Doc(en_vocab, words=["like"])])
    doc = Doc(en_vocab, words=["I", "like", "Google", "Now", "best"])
    assert "TEST" in matcher
    assert "TEST2" not in matcher
    assert len(matcher(doc)) == 1


def test_phrase_matcher_remove(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    matcher.add("TEST1", [Doc(en_vocab, words=["like"])])
    matcher.add("TEST2", [Doc(en_vocab, words=["best"])])
    doc = Doc(en_vocab, words=["I", "like", "Google", "Now", "best"])
    assert "TEST1" in matcher
    assert "TEST2" in matcher
    assert "TEST3" not in matcher
    assert len(matcher(doc)) == 2
    matcher.remove("TEST1")
    assert "TEST1" not in matcher
    assert "TEST2" in matcher
    assert "TEST3" not in matcher
    assert len(matcher(doc)) == 1
    matcher.remove("TEST2")
    assert "TEST1" not in matcher
    assert "TEST2" not in matcher
    assert "TEST3" not in matcher
    assert len(matcher(doc)) == 0
    with pytest.raises(KeyError):
        matcher.remove("TEST3")
    assert "TEST1" not in matcher
    assert "TEST2" not in matcher
    assert "TEST3" not in matcher
    assert len(matcher(doc)) == 0


def test_phrase_matcher_overlapping_with_remove(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    matcher.add("TEST", [Doc(en_vocab, words=["like"])])
    # TEST2 is added alongside TEST
    matcher.add("TEST2", [Doc(en_vocab, words=["like"])])
    doc = Doc(en_vocab, words=["I", "like", "Google", "Now", "best"])
    assert "TEST" in matcher
    assert len(matcher) == 2
    assert len(matcher(doc)) == 2
    # removing TEST does not remove the entry for TEST2
    matcher.remove("TEST")
    assert "TEST" not in matcher
    assert len(matcher) == 1
    assert len(matcher(doc)) == 1
    assert matcher(doc)[0][0] == en_vocab.strings["TEST2"]
    # removing TEST2 removes all
    matcher.remove("TEST2")
    assert "TEST2" not in matcher
    assert len(matcher) == 0
    assert len(matcher(doc)) == 0


def test_phrase_matcher_string_attrs(en_vocab):
    words1 = ["I", "like", "cats"]
    pos1 = ["PRON", "VERB", "NOUN"]
    words2 = ["Yes", ",", "you", "hate", "dogs", "very", "much"]
    pos2 = ["INTJ", "PUNCT", "PRON", "VERB", "NOUN", "ADV", "ADV"]
    pattern = Doc(en_vocab, words=words1, pos=pos1)
    matcher = PhraseMatcher(en_vocab, attr="POS")
    matcher.add("TEST", [pattern])
    doc = Doc(en_vocab, words=words2, pos=pos2)
    matches = matcher(doc)
    assert len(matches) == 1
    match_id, start, end = matches[0]
    assert match_id == en_vocab.strings["TEST"]
    assert start == 2
    assert end == 5


def test_phrase_matcher_string_attrs_negative(en_vocab):
    """Test that token with the control codes as ORTH are *not* matched."""
    words1 = ["I", "like", "cats"]
    pos1 = ["PRON", "VERB", "NOUN"]
    words2 = ["matcher:POS-PRON", "matcher:POS-VERB", "matcher:POS-NOUN"]
    pos2 = ["X", "X", "X"]
    pattern = Doc(en_vocab, words=words1, pos=pos1)
    matcher = PhraseMatcher(en_vocab, attr="POS")
    matcher.add("TEST", [pattern])
    doc = Doc(en_vocab, words=words2, pos=pos2)
    matches = matcher(doc)
    assert len(matches) == 0


def test_phrase_matcher_bool_attrs(en_vocab):
    words1 = ["Hello", "world", "!"]
    words2 = ["No", "problem", ",", "he", "said", "."]
    pattern = Doc(en_vocab, words=words1)
    matcher = PhraseMatcher(en_vocab, attr="IS_PUNCT")
    matcher.add("TEST", [pattern])
    doc = Doc(en_vocab, words=words2)
    matches = matcher(doc)
    assert len(matches) == 2
    match_id1, start1, end1 = matches[0]
    match_id2, start2, end2 = matches[1]
    assert match_id1 == en_vocab.strings["TEST"]
    assert match_id2 == en_vocab.strings["TEST"]
    assert start1 == 0
    assert end1 == 3
    assert start2 == 3
    assert end2 == 6


def test_phrase_matcher_validation(en_vocab):
    doc1 = Doc(en_vocab, words=["Test"])
    doc1[0].dep_ = "ROOT"
    doc2 = Doc(en_vocab, words=["Test"])
    doc2[0].tag_ = "TAG"
    doc2[0].pos_ = "X"
    doc2[0].set_morph("Feat=Val")
    doc3 = Doc(en_vocab, words=["Test"])
    matcher = PhraseMatcher(en_vocab, validate=True)
    with pytest.warns(UserWarning):
        matcher.add("TEST1", [doc1])
    with pytest.warns(UserWarning):
        matcher.add("TEST2", [doc2])
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        matcher.add("TEST3", [doc3])
    matcher = PhraseMatcher(en_vocab, attr="POS", validate=True)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        matcher.add("TEST4", [doc2])


def test_attr_validation(en_vocab):
    with pytest.raises(ValueError):
        PhraseMatcher(en_vocab, attr="UNSUPPORTED")


def test_attr_pipeline_checks(en_vocab):
    doc1 = Doc(en_vocab, words=["Test"])
    doc1[0].dep_ = "ROOT"
    doc2 = Doc(en_vocab, words=["Test"])
    doc2[0].tag_ = "TAG"
    doc2[0].pos_ = "X"
    doc2[0].set_morph("Feat=Val")
    doc2[0].lemma_ = "LEMMA"
    doc3 = Doc(en_vocab, words=["Test"])
    # DEP requires DEP
    matcher = PhraseMatcher(en_vocab, attr="DEP")
    matcher.add("TEST1", [doc1])
    with pytest.raises(ValueError):
        matcher.add("TEST2", [doc2])
    with pytest.raises(ValueError):
        matcher.add("TEST3", [doc3])
    # TAG, POS, LEMMA require those values
    for attr in ("TAG", "POS", "LEMMA"):
        matcher = PhraseMatcher(en_vocab, attr=attr)
        matcher.add("TEST2", [doc2])
        with pytest.raises(ValueError):
            matcher.add("TEST1", [doc1])
        with pytest.raises(ValueError):
            matcher.add("TEST3", [doc3])
    # TEXT/ORTH only require tokens
    matcher = PhraseMatcher(en_vocab, attr="ORTH")
    matcher.add("TEST3", [doc3])
    matcher = PhraseMatcher(en_vocab, attr="TEXT")
    matcher.add("TEST3", [doc3])


def test_phrase_matcher_callback(en_vocab):
    mock = Mock()
    doc = Doc(en_vocab, words=["I", "like", "Google", "Now", "best"])
    pattern = Doc(en_vocab, words=["Google", "Now"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("COMPANY", [pattern], on_match=mock)
    matches = matcher(doc)
    mock.assert_called_once_with(matcher, doc, 0, matches)


def test_phrase_matcher_remove_overlapping_patterns(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    pattern1 = Doc(en_vocab, words=["this"])
    pattern2 = Doc(en_vocab, words=["this", "is"])
    pattern3 = Doc(en_vocab, words=["this", "is", "a"])
    pattern4 = Doc(en_vocab, words=["this", "is", "a", "word"])
    matcher.add("THIS", [pattern1, pattern2, pattern3, pattern4])
    matcher.remove("THIS")


def test_phrase_matcher_basic_check(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    # Potential mistake: pass in pattern instead of list of patterns
    pattern = Doc(en_vocab, words=["hello", "world"])
    with pytest.raises(ValueError):
        matcher.add("TEST", pattern)


def test_phrase_matcher_pickle(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    mock = Mock()
    matcher.add("TEST", [Doc(en_vocab, words=["test"])])
    matcher.add("TEST2", [Doc(en_vocab, words=["test2"])], on_match=mock)
    doc = Doc(en_vocab, words=["these", "are", "tests", ":", "test", "test2"])
    assert len(matcher) == 2

    b = srsly.pickle_dumps(matcher)
    matcher_unpickled = srsly.pickle_loads(b)

    # call after pickling to avoid recursion error related to mock
    matches = matcher(doc)
    matches_unpickled = matcher_unpickled(doc)

    assert len(matcher) == len(matcher_unpickled)
    assert matches == matches_unpickled

    # clunky way to vaguely check that callback is unpickled
    (vocab, docs, callbacks, attr) = matcher_unpickled.__reduce__()[1]
    assert isinstance(callbacks.get("TEST2"), Mock)


def test_phrase_matcher_as_spans(en_vocab):
    """Test the new as_spans=True API."""
    matcher = PhraseMatcher(en_vocab)
    matcher.add("A", [Doc(en_vocab, words=["hello", "world"])])
    matcher.add("B", [Doc(en_vocab, words=["test"])])
    doc = Doc(en_vocab, words=["...", "hello", "world", "this", "is", "a", "test"])
    matches = matcher(doc, as_spans=True)
    assert len(matches) == 2
    assert isinstance(matches[0], Span)
    assert matches[0].text == "hello world"
    assert matches[0].label_ == "A"
    assert isinstance(matches[1], Span)
    assert matches[1].text == "test"
    assert matches[1].label_ == "B"


def test_phrase_matcher_deprecated(en_vocab):
    matcher = PhraseMatcher(en_vocab)
    matcher.add("TEST", [Doc(en_vocab, words=["helllo"])])
    doc = Doc(en_vocab, words=["hello", "world"])
    with pytest.warns(DeprecationWarning) as record:
        for _ in matcher.pipe([doc]):
            pass
        assert record.list
        assert "spaCy v3.0" in str(record.list[0].message)


@pytest.mark.parametrize("attr", ["SENT_START", "IS_SENT_START"])
def test_phrase_matcher_sent_start(en_vocab, attr):
    _ = PhraseMatcher(en_vocab, attr=attr)  # noqa: F841


def test_span_in_phrasematcher(en_vocab):
    """Ensure that PhraseMatcher accepts Span and Doc as input"""
    # fmt: off
    words = ["I", "like", "Spans", "and", "Docs", "in", "my", "input", ",", "and", "nothing", "else", "."]
    # fmt: on
    doc = Doc(en_vocab, words=words)
    span = doc[:8]
    pattern = Doc(en_vocab, words=["Spans", "and", "Docs"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("SPACY", [pattern])
    matches_doc = matcher(doc)
    matches_span = matcher(span)
    assert len(matches_doc) == 1
    assert len(matches_span) == 1


def test_span_v_doc_in_phrasematcher(en_vocab):
    """Ensure that PhraseMatcher only returns matches in input Span and not in entire Doc"""
    # fmt: off
    words = [
        "I", "like", "Spans", "and", "Docs", "in", "my", "input", ",", "Spans",
        "and", "Docs", "in", "my", "matchers", "," "and", "Spans", "and", "Docs",
        "everywhere", "."
    ]
    # fmt: on
    doc = Doc(en_vocab, words=words)
    span = doc[9:15]  # second clause
    pattern = Doc(en_vocab, words=["Spans", "and", "Docs"])
    matcher = PhraseMatcher(en_vocab)
    matcher.add("SPACY", [pattern])
    matches_doc = matcher(doc)
    matches_span = matcher(span)
    assert len(matches_doc) == 3
    assert len(matches_span) == 1



from cymem.cymem cimport Pool
from thinc.backends.cblas cimport CBlas

from ..ml.parser_model cimport ActivationsC, SizesC, WeightsC
from ..vocab cimport Vocab
from ._parser_internals._state cimport StateC
from ._parser_internals.transition_system cimport Transition, TransitionSystem
from .trainable_pipe cimport TrainablePipe


cdef class Parser(TrainablePipe):
    cdef public object _rehearsal_model
    cdef readonly TransitionSystem moves
    cdef public object _multitasks

    cdef void _parseC(
        self,
        CBlas cblas,
        StateC** states,
        WeightsC weights,
        SizesC sizes
    ) nogil

    cdef void c_transition_batch(
        self,
        StateC** states,
        const float* scores,
        int nr_class,
        int batch_size
    ) nogil



from dataclasses import dataclass
from typing import List

from .align import get_alignments
from .alignment_array import AlignmentArray


@dataclass
class Alignment:
    x2y: AlignmentArray
    y2x: AlignmentArray

    @classmethod
    def from_indices(cls, x2y: List[List[int]], y2x: List[List[int]]) -> "Alignment":
        x2y = AlignmentArray(x2y)
        y2x = AlignmentArray(y2x)
        return Alignment(x2y=x2y, y2x=y2x)

    @classmethod
    def from_strings(cls, A: List[str], B: List[str]) -> "Alignment":
        x2y, y2x = get_alignments(A, B)
        return Alignment.from_indices(x2y=x2y, y2x=y2x)



"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.gu.examples import sentences
>>> docs = nlp.pipe(sentences)
"""


sentences = [
    "લોકશાહી એ સરકારનું એક એવું તંત્ર છે જ્યાં નાગરિકો મત દ્વારા સત્તાનો ઉપયોગ કરે છે.",
    "તે ગુજરાત રાજ્યના ધરમપુર શહેરમાં આવેલું હતું",
    "કર્ણદેવ પહેલો સોલંકી વંશનો રાજા હતો",
    "તેજપાળને બે પત્ની હતી",
    "ગુજરાતમાં ભારતીય જનતા પક્ષનો ઉદય આ સમયગાળા દરમિયાન થયો",
    "આંદોલનકારીઓએ ચીમનભાઇ પટેલના રાજીનામાની માંગણી કરી.",
    "અહિયાં શું જોડાય છે?",
    "મંદિરનો પૂર્વાભિમુખ ભાગ નાના મંડપ સાથે થોડો લંબચોરસ આકારનો છે.",
]
